# Author: Adrián Tóth <adtoth@redhat.com>
#
# Copyright (c) 2020, Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
import re

from ansiblelater.standard import StandardBase


class CheckNestedJinja(StandardBase):

    sid = "ANSIBLE0023"
    description = "Don't use nested Jinja2 pattern"
    helptext = (
        "there should not be any nested jinja pattern "
        "like `{{ list_one + {{ list_two | max }} }}`"
    )
    version = "0.2"
    types = ["playbook", "task", "handler", "rolevars", "hostvars", "groupvars"]

    def check(self, candidate, settings):
        yamllines, errors = self.get_normalized_yaml(candidate, settings)
        pattern = re.compile(r"{{(?:[^{}]*)?[^'\"]{{")
        matches = []

        if not errors:
            for i, line in yamllines:
                if "!unsafe" in line:
                    continue

                match = pattern.findall(line)
                if match:
                    for item in match:
                        matches.append((i, item))

            for i, _ in matches:
                errors.append(self.Error(i, self.helptext))

        return self.Result(candidate.path, errors)



import re
from collections import defaultdict

from ansiblelater.standard import StandardBase


class CheckTaskSeparation(StandardBase):

    sid = "ANSIBLE0001"
    description = "Single tasks should be separated by empty line"
    helptext = "missing task separation (required: 1 empty line)"
    version = "0.1"
    types = ["playbook", "task", "handler"]

    def check(self, candidate, settings):
        options = defaultdict(dict)
        options.update(remove_empty=False)
        options.update(remove_markers=False)

        yamllines, line_errors = self.get_normalized_yaml(candidate, settings, options)
        tasks, task_errors = self.get_normalized_tasks(candidate, settings)

        task_regex = re.compile(r"-\sname:(.*)")
        prevline = "#file_start_marker"

        allowed_prevline = [
            "---",
            "handlers:",
            "tasks:",
            "pre_tasks:",
            "post_tasks:",
            "block:",
            "rescue:",
            "always:",
        ]

        errors = task_errors + line_errors
        if not errors:
            for i, line in yamllines:
                match = task_regex.search(line)
                if match and prevline:
                    name = match.group(1).strip()

                    if not any(task.get("name") == name for task in tasks):
                        continue

                    if not any(item in prevline for item in allowed_prevline):
                        errors.append(self.Error(i, self.helptext))

                prevline = line.strip()

        return self.Result(candidate.path, errors)



import os

from ansiblelater.standard import StandardBase


class CheckYamlFile(StandardBase):

    sid = "LINT0006"
    description = "Roles file should be in yaml format"
    helptext = "file does not have a .yml extension"
    version = "0.1"
    types = ["playbook", "task", "handler"]

    def check(self, candidate, settings):
        errors = []
        extensions = [".yml", ".yaml"]

        if os.path.isfile(candidate.path) and os.path.splitext(candidate.path)[1] in extensions:
            content, errors = self.get_raw_yaml(candidate, settings)
        else:
            errors.append(self.Error(None, self.helptext))

        return self.Result(candidate.path, errors)



import re

from ansiblelater.standard import StandardBase


class CheckFilterSeparation(StandardBase):

    sid = "ANSIBLE0016"
    description = "Jinja2 filters should be separated with spaces"
    helptext = "no suitable numbers of spaces (required: 1)"
    version = "0.1"
    types = ["playbook", "task", "handler", "rolevars", "hostvars", "groupvars"]

    def check(self, candidate, settings):
        yamllines, errors = self.get_normalized_yaml(candidate, settings)

        matches = []
        braces = re.compile("{{(.*?)}}")
        filters = re.compile(r"(?<=\|)((\s{2,})*\S+)|(\S+(\s{2,})*)(?=\|)")

        if not errors:
            for i, line in yamllines:
                match = braces.findall(line)
                if match:
                    for item in match:
                        # replace potential regex in filters
                        item = re.sub(r"\(.+\)", "(dummy)", item)
                        matches.append((i, item))

            for i, item in matches:
                if filters.findall(item):
                    errors.append(self.Error(i, self.helptext))
        return self.Result(candidate.path, errors)



from ansiblelater.standard import StandardBase


class CheckDeprecated(StandardBase):

    sid = "ANSIBLE9999"
    description = "Deprecated features should not be used"
    helptext = "'{old}' is deprecated and should not be used anymore. Use '{new}' instead."
    version = "0.1"
    types = ["playbook", "task", "handler"]

    def check(self, candidate, settings):
        tasks, errors = self.get_normalized_tasks(candidate, settings, full=True)

        if not errors:
            for task in tasks:
                if "skip_ansible_lint" in (task.get("tags") or []):
                    errors.append(
                        self.Error(
                            task["__line__"],
                            self.helptext.format(
                                old="skip_ansible_lint", new="skip_ansible_later"
                            )
                        )
                    )
        return self.Result(candidate.path, errors)



from ansiblelater.standard import StandardBase


class CheckYamlHasContent(StandardBase):

    sid = "LINT0007"
    description = "Files should contain useful content"
    helptext = "the file appears to have no useful content"
    version = "0.1"
    types = ["playbook", "task", "handler", "rolevars", "defaults", "meta"]

    def check(self, candidate, settings):
        yamllines, errors = self.get_normalized_yaml(candidate, settings)

        if (not candidate.faulty and len(yamllines) == 0) and not errors:
            errors.append(self.Error(None, self.helptext))

        return self.Result(candidate.path, errors)



from ansiblelater.standard import StandardBase


class CheckShellInsteadCommand(StandardBase):

    sid = "ANSIBLE0010"
    description = "Shell should only be used when essential"
    helptext = "shell should only be used when piping, redirecting or chaining commands"
    version = "0.1"
    types = ["playbook", "task", "handler"]

    def check(self, candidate, settings):
        tasks, errors = self.get_normalized_tasks(candidate, settings)

        if not errors:
            for task in tasks:
                if task["action"]["__ansible_module__"] == "shell":
                    # skip processing if args.executable is used as this
                    # parameter is no longer support by command module
                    if "executable" in task["action"]:
                        continue

                    cmd = self.get_safe_cmd(task)
                    if not any(ch in cmd for ch in self.SHELL_PIPE_CHARS):
                        errors.append(self.Error(task["__line__"], self.helptext))

        return self.Result(candidate.path, errors)



from ansible.parsing.yaml.objects import AnsibleMapping

from ansiblelater.standard import StandardBase


class CheckScmInSrc(StandardBase):

    sid = "ANSIBLE0005"
    description = "Use `scm:` key rather than `src: scm+url`"
    helptext = "usage of `src: scm+url` not recommended"
    version = "0.1"
    types = ["rolesfile"]

    def check(self, candidate, settings):
        roles, errors = self.get_tasks(candidate, settings)

        if not errors:
            for role in roles:
                if (
                    isinstance(role, AnsibleMapping) and bool(role.get("src"))
                    and "+" in role.get("src")
                ):
                    errors.append(self.Error(role["__line__"], self.helptext))

        return self.Result(candidate.path, errors)



"""Test settings module."""

import pytest

from ansiblelater import settings


@pytest.fixture
def settings_instance():
    c = settings.Settings(args={"rules": {"files": []}})

    return c


def test_args_member(settings_instance):
    x = {"rules": {"files": ["*"]}}

    assert x == settings_instance.args


def test_args_setter(settings_instance):
    default = {"rules.files": ["dummy"], "config_file": "conf.yml"}
    x = {"rules": {"files": ["dummy"]}}

    s = settings_instance._set_args(default)

    assert x == s



from ansiblelater.standard import StandardBase


class CheckVersion(StandardBase):

    sid = "ANSIBLE9998"
    description = "Standards version should be pinned"
    helptext = "Standards version not set. Using latest standards version {version}"
    types = ["task", "handler", "rolevars", "meta", "template", "file", "playbook"]

    def check(self, candidate, settings):  # noqa
        errors = []

        if not candidate.version_config:
            errors.append(self.Error(None, self.helptext.format(version=candidate.version)))

        return self.Result(candidate.path, errors)



from ansiblelater.standard import StandardBase


class CheckBecomeUser(StandardBase):

    sid = "ANSIBLE0015"
    description = "Become should be combined with become_user"
    helptext = "the task has `become` enabled but `become_user` is missing"
    version = "0.1"
    types = ["playbook", "task", "handler"]

    def check(self, candidate, settings):
        tasks, errors = self.get_normalized_tasks(candidate, settings)
        true_value = [True, "true", "True", "TRUE", "yes", "Yes", "YES"]

        if not errors:
            gen = (task for task in tasks if "become" in task)
            for task in gen:
                if task["become"] in true_value and "become_user" not in task.keys():
                    errors.append(self.Error(task["__line__"], self.helptext))

        return self.Result(candidate.path, errors)



from ansiblelater.standard import StandardBase


class CheckWhenFormat(StandardBase):

    sid = "ANSIBLE0022"
    description = "Don't use Jinja2 in when"
    helptext = (
        "`when` is a raw Jinja2 expression, redundant {{ }} "
        "should be removed from variable(s)"
    )
    version = "0.2"
    types = ["playbook", "task", "handler"]

    def check(self, candidate, settings):
        tasks, errors = self.get_normalized_tasks(candidate, settings)

        if not errors:
            for task in tasks:
                if "when" in task and not self._is_valid(task["when"]):
                    errors.append(self.Error(task["__line__"], self.helptext))

        return self.Result(candidate.path, errors)

    @staticmethod
    def _is_valid(when):
        if not isinstance(when, str):
            return True
        return when.find("{{") == -1 and when.find("}}") == -1



# Copyright (c) 2013-2014 Will Thames <will@thames.id.au>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import os

from ansiblelater.standard import StandardBase


class CheckCommandInsteadOfArgument(StandardBase):

    sid = "ANSIBLE0017"
    description = "Commands should not be used in place of module arguments"
    helptext = "{exec} used in place of file modules argument {arg}"
    version = "0.2"
    types = ["playbook", "task", "handler"]

    def check(self, candidate, settings):
        tasks, errors = self.get_normalized_tasks(candidate, settings)
        commands = ["command", "shell", "raw"]
        arguments = {
            "chown": "owner",
            "chmod": "mode",
            "chgrp": "group",
            "ln": "state=link",
            "mkdir": "state=directory",
            "rmdir": "state=absent",
            "rm": "state=absent"
        }

        if not errors:
            for task in tasks:
                if task["action"]["__ansible_module__"] in commands:
                    first_cmd_arg = self.get_first_cmd_arg(task)
                    executable = os.path.basename(first_cmd_arg)

                    if (
                        first_cmd_arg and executable in arguments
                        and task["action"].get("warn", True)
                    ):
                        errors.append(
                            self.Error(
                                task["__line__"],
                                self.helptext.format(exec=executable, arg=arguments[executable])
                            )
                        )

        return self.Result(candidate.path, errors)



from ansiblelater.standard import StandardBase


class CheckNamedTask(StandardBase):

    sid = "ANSIBLE0006"
    description = "Tasks and handlers must be named"
    helptext = "module '{module}' used without or empty `name` attribute"
    version = "0.1"
    types = ["playbook", "task", "handler"]

    def check(self, candidate, settings):
        tasks, errors = self.get_normalized_tasks(candidate, settings)
        exclude_modules = settings["ansible"]["named-task"]["exclude"]

        if not errors:
            for task in tasks:
                module = task["action"]["__ansible_module__"]
                if ("name" not in task or not task["name"]) and module not in exclude_modules:
                    errors.append(
                        self.Error(task["__line__"], self.helptext.format(module=module))
                    )

        return self.Result(candidate.path, errors)



import re

from ansiblelater.standard import StandardBase
from ansiblelater.utils import count_spaces


class CheckBracesSpaces(StandardBase):

    sid = "ANSIBLE0004"
    description = "YAML should use consistent number of spaces around variables"
    helptext = "no suitable numbers of spaces (min: {min} max: {max})"
    version = "0.1"
    types = ["playbook", "task", "handler", "rolevars", "hostvars", "groupvars", "meta"]

    def check(self, candidate, settings):
        yamllines, errors = self.get_normalized_yaml(candidate, settings)
        conf = settings["ansible"]["double-braces"]

        matches = []
        braces = re.compile("{{(.*?)}}")

        if not errors:
            for i, line in yamllines:
                if "!unsafe" in line:
                    continue
                match = braces.findall(line)
                if match:
                    for item in match:
                        matches.append((i, item))

            for i, line in matches:
                [leading, trailing] = count_spaces(line)
                sum_spaces = leading + trailing

                if (
                    sum_spaces < conf["min-spaces-inside"] * 2
                    or sum_spaces > conf["min-spaces-inside"] * 2
                ):
                    errors.append(
                        self.Error(
                            i,
                            self.helptext.format(
                                min=conf["min-spaces-inside"], max=conf["max-spaces-inside"]
                            )
                        )
                    )
        return self.Result(candidate.path, errors)



import re

from ansiblelater.candidate import Template
from ansiblelater.standard import StandardBase


class CheckCompareToEmptyString(StandardBase):

    sid = "ANSIBLE0012"
    description = "Don't compare to empty string \"\""
    helptext = ("use `when: var` rather than `when: var !=` (or conversely `when: not var`)")
    version = "0.1"
    types = ["playbook", "task", "handler", "template"]

    def check(self, candidate, settings):
        yamllines, errors = self.get_normalized_yaml(candidate, settings)

        if not errors:
            if isinstance(candidate, Template):
                matches = []
                jinja_string = re.compile("({{|{%)(.*?)(}}|%})")

                for i, line in yamllines:
                    match = jinja_string.findall(line)
                    if match:
                        for item in match:
                            matches.append((i, item[1]))

                yamllines = matches

            empty_string_compare = re.compile("[=!]= ?[\"'][\"']")

            for i, line in yamllines:
                if empty_string_compare.findall(line):
                    errors.append(self.Error(i, self.helptext))

        return self.Result(candidate.path, errors)
